#define M1 4
#define E1 5
#define M2 7
#define E2 6

#define PORT_BUTTON 0 // Cоздаём макроопределение с номером портоа А0 для считывания нажатия кнопки 

//Создаём макроопределение для хранения константы напряжения на единицу значения
#define CONST_U 0.02355 // 0.024414 - расчётное значение

//Создаём макроопределение для хранения номера аналогового порта с подключённым к нему датчиком напряжения А11(D12)
#define PORT_U 11

//Создаём макроопределение для хранения номера цифрового порта с подключённым к нему светодиодом
#define LED_PIN 13

//Создаём макроопределение для хранения значения минимального напряжения на аккумуляторе
// ниже которого произойдёт отключение робота во избежание переразряда аккумулятора 
#define MIN_U 5.9

// Блок деректив препроцессора для создания макроопределений содержащих коэффициенты для расчёта управляющего
// воздействия в формуле PD регулятора  
#define Pk 3.5    // Коэффициент для пропорциональной составляющей 3.5
#define Dk 5.0  // Коэффициент для дифференциальной составляющей 5.0


#define MS 50 // Макроопределение содержащее исходную скорость движения робота 50 


//Объявление массива типа int с номерами аналоговых портов на МК
int port[8] = {1, 2, 3, 4, 5, 8, 9, 10};

//Объявление массива типа byte из 8 элементов в котором будут содержатся значения датчиков 0 или 1
byte zn_datch[8];

//Объявления массива с весовыми коэффициентами
int ves[8] = { -40, -15, -7, 0, 0, 7, 15, 40};

//Объявление переменных
int U;      // Переменная содержащая значение управляющего воздействия
int U_old;  // Переменная содержащая значение предыдущего управляющего воздействия

long SensorsOpt = 0;    // Переменная содержащая оптимальное значение датчиков (робот точно по центру линии
long SensorsValue = 0;  // Переменная содержащая текущее значение датчиков (робот точно по центру линии
 
int Err = 0;    // Переменная содержащая ошибку
int ErrOld = 0; // Переменная содержащая ошибку на предыдущем шаге
int p;          // Переменная содержащая пропорциональную составляющую формулы PD регулятора
int d;          // Переменная содержащая дифференциальную составляющую формулы PD регулятора
int Sum = 0;    // Переменная содержащая сумму всех датчиков

float K=1.0; // Переменная содержащая поправочный коэффициент


// Обязательная функция программы выполняется только один раз при включении микроконтроллерного устройства
// используется для инициализации сериального порта для обмена данными с компьютером и для первичной конфигурации портов и т.д.
void setup()
{
  Serial.begin(9600);    // Открываем сериальный порт для обмена данными на скорости 9600 бод (бит/с)

  // Конфигурируем порты микроконтроллера для управления двигателями как выходные
  pinMode(M1, OUTPUT);
  pinMode(E1, OUTPUT);
  pinMode(M2, OUTPUT);
  pinMode(E2, OUTPUT);
  delay(100); // Делаем задержку в выполнении программы 100мс (0,1 сек.)

  // Конфигурируем порт А0 с подключенным к нему блоком кнопок на получение данных, т.е. на вход
  pinMode(PORT_BUTTON, INPUT);

  // Конфигурируем порт микроконтроллера для получения данных с датчика напряжения
  pinMode(PORT_U, INPUT);

  // Конфигурируем порт микроконтроллера c подключённым к нему светодиодом, на выход
  pinMode(LED_PIN, OUTPUT);

  while (button()!=2);

  
  readDat(); //Первоначальное чтение датчиков для заполнения Оптимального значения датчиков т.е. линия ровно посредине
  
  // Цикл в котором мы получаем оптимальное значение датчиков, это делается при старте робота
  // когда он размещён точно по центру линии
  for (int i = 0; i < 8; i++)
  {
    // Значение получается путём умножения значения датчика на соответствующий ему весовой коэффициент
    SensorsOpt += (zn_datch[i] * ves[i]); 
  }

}

// Обязательная функция программы выполняется как бесконечный цикл
void loop()
{
  //Вызов пользовательской функции datch_u() для определения напряжения на аккумуляторе
  datch_u();

  //Чтение датчиков
  readDat();


  /////////////////////////////////////////////////////////////////////////////////////
  // Отслеживаем потерю линии т.е. все датчики белые, если все датчики на белом,
  // мы не попадём в этот оператор, значит не будем рассчитывать управляющее воздействие,
  // и будем пользоваться последним рассчитанным управляющим воздействием когда 
  // над чёрной линией был крайний датчик и соответственно управляющее воздействие
  // было максимальным
  if (Sum != 0)
  {
    
    SensorsValue = 0;
    for (int i = 0; i < 8; i++)
    {
      SensorsValue += (zn_datch[i] * ves[i]);
    }

    //Вычисление текущей ошибки
    Err = SensorsValue - SensorsOpt;
   
    //Расчет Пропорциональной составляющей
    p = Pk * Err;

    //Расчет Дифференциальной составляющей
    d = Dk * (Err - ErrOld);

    //Расчет управляющего воздействия
    U = p + d;

    //Присваиваем значение текущей ошибки переменной старой ошибки
    ErrOld = Err;
  }
  /////////////////////////////////////////////////////////////////////////////////////
  

  /////////////////////////////////////////////////////////////////////////////////////
  //Фильтр для проверки значения управляющего воздействия
  if (U < 0)  //Если управляющее воздействие отрицательное проверяем знак показаний датчиков
  {
    if (SensorsValue > 0) //Если знак показаний датчиков не совпадает со знаком управляющего воздействия
                          //считаем его "шумом" и присваиваем предыдущее значение
    { 
      U = U_old;
    }
  }
  else if (U > 0)
  {
    if (SensorsValue < 0) //Если знак показаний датчиков не совпадает со знаком управляющего воздействия
                          //считаем его "шумом" и присваиваем предыдущее значение
    { 
      U = U_old;
    }
  }
  /////////////////////////////////////////////////////////////////////////////////////

  U *= K; // Умножаем значение управляющего воздействия на коэффициент поправки
  
  Motor(MS - U, MS + U); //передаем управляющее воздействие на моторы
  
  U_old = U; //Сохраняем предыдущее значение управляющего воздействия

}
